// define this to run in 8 bpp mode; the XBuf -> ptbuffer copy loop is marginally faster in 8 bpp
#define BPP8

#define PROFILE

#include <gpstdio.h>
#include <gpgraphic.h>
#include <gpfont.h>
#include <gpmem.h>
#include <gpstdlib.h>
#include <string.h>
#include "Coleco.h"
#define BOOL char
#include "gpfilereq.h"
#include "gp32.h"
GPDRAWSURFACE gpDraw[2];
int gpDrawFlip=0;

#include <gpstd.h>

#ifdef BPP8
	#define print print8
#else
	#define print print16
#endif

#include "gpsoundbuf.h"

GPSOUNDBUF sb;

void SoundCallback(void* userdata, unsigned char* stream, int len);

int InitMachine(void)
{
	int i;
	GpKeyInit();
#ifndef BPP8
	GpGraphicModeSet(GPC_DFLAG_16BPP,NULL);
#else /* BPP8 */
	GpGraphicModeSet(GPC_DFLAG_8BPP,NULL);
#endif /* BPP8 */
	for(i=0;i<2;i++)
	{
		GpLcdSurfaceGet(&gpDraw[i], i);
	}
	GpPcmInit(PCM_M44,PCM_8BIT);
	sb.freq=PCM_M44;
	sb.format=PCM_8BIT;
	sb.samples=44100/40;
	sb.callback=SoundCallback;
	sb.pollfreq=240;
	GpSoundBufStart(&sb);
	return 0;
}

void TrashMachine(void)
{
	GpSoundBufStop();
}

#define WIDTH 320
#define HEIGHT 240
byte XBuf[WIDTH*HEIGHT];
byte XPal[16];
byte XPal0=0;
void PutImage(void);
#include "Common.h"
#undef RefreshScreen

unsigned short colors[256];

void SetColor(byte N, byte R, byte B, byte G)
{
#ifndef BPP8
	colors[N]=((R>>3)<<11)+((B>>3)<<6)+((G>>3)<<1);
#else /* BPP8 */
	unsigned short color;
	color=((R>>3)<<11)+((B>>3)<<6)+((G>>3)<<1);
	GpPaletteEntryChange(N,1,&color,0);
#endif /* BPP8 */
	XPal[N]=N;
}

int freqs[8]={0};
int vols[8]={0};
static int Count[8]={0};
#define SND_BUFSIZE 128
static int Wave[SND_BUFSIZE];
static int NoiseGen=1;
#define AUDIO_CONV(A) (128+(A))
static int MasterVolume=192;

// audio ring buffer
#define RBUF_SIZE 2048
unsigned char RingBuf[RBUF_SIZE];
volatile int rbwpos=0;
volatile int rbrpos=0;
volatile int rbfill=0;

int count=0;

int keypad=-1;
void PutImage(void)
{
	static int last=-1;
	static int now1;
#ifdef PROFILE
	static int now2;
#endif
	static char buf[20];
	int i,j;
#ifndef BPP8
	unsigned short* fb;
#else /* BPP8 */
	unsigned char* fb;
#endif /* BPP8 */
	count++;
	if(last==-1) last=GpTickCountGet();
	now1=GpTickCountGet()-last;
	fb=gpDraw[gpDrawFlip].ptbuffer;


	for(i=0;i<HEIGHT;i++)
		for(j=0;j<WIDTH;j++)
#ifndef BPP8
			fb[j*HEIGHT+(HEIGHT-1-i)]=colors[XBuf[i*WIDTH+j]];
#else /* BPP8 */
			fb[j*HEIGHT+(HEIGHT-1-i)]=XBuf[i*WIDTH+j];
#endif /* BPP8 */
	if(keypad!=-1)
	{
		strcpy(buf,"KEY x");
		if(keypad<10) buf[4]=keypad+'0';
		else if(keypad==10)	buf[4]='#';
		else if(keypad==11)	buf[4]='*';
#ifdef BPP8
		GpRectFill(0,&gpDraw[gpDrawFlip],320-5*8,0,320,16,0);
		GpTextOut(0,&gpDraw[gpDrawFlip],320-5*8,0,buf,0xfe);
#else
		print(320/8-5,0,buf);
#endif /* BPP8 */
	}
#ifdef PROFILE
	now2=GpTickCountGet()-now1-last;
	for(i=0;i<now1;i++)
	{
#ifdef BPP8
		fb[i]=0xfe;
		fb[i+240]=0xfe;
		fb[i+480]=0xfe;
#else
		fb[i]=0xfffe;
		fb[i+240]=0xfffe;
		fb[i+480]=0xfffe;
#endif
	}
	for(i=120;i<now2+120;i++)
	{
#ifdef BPP8
		fb[i]=0xee;
		fb[i+240]=0xee;
		fb[i+480]=0xee;
#else
		fb[i]=0xffee;
		fb[i+240]=0xffee;
		fb[i+480]=0xffee;
#endif
	}
#ifdef BPP8
	fb[16]=0x34;
	fb[32]=0x34;
	fb[48]=0x34;
#else
	fb[16]=0x1234;
	fb[32]=0x1234;
	fb[48]=0x1234;
#endif
#endif // PROFILE

	// Audio ringbuffer fill bar
	//gm_memset(fb+480, 0xfe, rbfill/16);

	GpSurfaceFlip(&gpDraw[gpDrawFlip]);
	gpDrawFlip=(gpDrawFlip+1)&1;

	/* Waveform generator */
	int V,I,J,K,L1,L2,M;
	for(M=0;M<=44100/SND_BUFSIZE/(60/UPeriod);M++)
	{
		for(J=0;J<8;J++)
			if(freqs[J]&&(V=vols[J]))
			{
				if(J!=3)
				{
					/* Melodic Sound */
					K=0x10000*freqs[J]/44100;
					L1=Count[J];
					for(I=0;I<SND_BUFSIZE;I++)
					{
						L2=L1+K;
						Wave[I]+=
						V*(L1&0x8000? (L2&0x8000? 32767:0):(L2&0x8000? 0:-32767));
						L1=L2;
					}
					Count[J]=L1;
				}
				else
				{

					/* White Noise */
					K=0x10000*freqs[J]/44100;
					L1=Count[J];
					for(I=0;I<SND_BUFSIZE;I++)
					{
						L1+=K;
						if(L1&0xFFFF0000)
						{
							L1&=0xFFFF;
							NoiseGen<<=1;
							if(NoiseGen&0x80000000)	NoiseGen^=0x08000001;
						}
						Wave[I]+=V*(NoiseGen&1? 32767:-32767);
					}
					Count[J]=L1;
				}
			}

			/* Mix and convert waveforms */
		for(J=0;J<SND_BUFSIZE;J++)
		{
			RingBuf[rbwpos]=AUDIO_CONV((Wave[J]/8)*MasterVolume/256/256/256);
			Wave[J]=0;
			rbwpos=(rbwpos+1)&(RBUF_SIZE-1);
			rbfill++;
			while(rbfill>RBUF_SIZE); // wait if buffer full
		}
	}

#if 0
	if(GpTickCountGet()-last<33*UPeriod)
	{
#ifdef PROFILE
		fb[0]=fb[240]=fb[1]=fb[241]=
#ifdef BPP8
							0x20;
#else
							0x4320;
#endif
#endif // PROFILE
		while(GpTickCountGet()-last<33*UPeriod);
	}
#endif
	last=GpTickCountGet();
}

extern byte ExitNow;

void Joysticks(void)
{
	static int endkeypad=0;
	static int fl=0;

	int keys;
	JoyState[1]=0xffff;
	GpKeyGetEx(&keys);
	JoyState[0]=0x0;
#define SETKEY(gpkey,shift) if(keys&(gpkey)) JoyState[0]+=1<<(shift)
	SETKEY(GPC_VK_FA,6);
	SETKEY(GPC_VK_UP,8);
	SETKEY(GPC_VK_RIGHT,9);
	SETKEY(GPC_VK_DOWN,10);
	SETKEY(GPC_VK_LEFT,11);
	SETKEY(GPC_VK_FB,14);
	if(keys&GPC_VK_FL)
	{
		if(!fl)
		{
			keypad=(keypad+1)%12;
			fl=1;
		}
	}
	else fl=0;

	if(keys&GPC_VK_FR && keypad!=-1)	// fire a keypad press
	{
		JoyState[0]+=(~keypad)&0xf;
		endkeypad=1;
	}
	if(!(keys&GPC_VK_FR) && endkeypad)
	{
		endkeypad=0;
		keypad=-1;
	}
	JoyState[0]=~JoyState[0];
	if(keys&GPC_VK_START) ExitNow=1;
}

extern byte Verbose;

void GpMain(void* arg)
{
	//GpClockSpeedChange (67500000, 0x25002, 1);
	GpClockSpeedChange (132000000, 0x3a011, 3);
	//EnableCache();
	int error;
	char file[12];
	FSEL f;
	GpGraphicModeSet(GPC_DFLAG_8BPP,NULL);
	GpLcdSurfaceGet(&gpDraw[0], 0);
	error=GpFatInit(); if(error) printf(0,0,"fatinit error");
	f.fname=(char*)file;
	f.gpDraw=&gpDraw[0];
	f.bgcolor=0xff;
	f.dircolor=0x2;
	f.filecolor=0;
	f.dirselcolor=0x3;
	f.fileselcolor=0x4a;
	f.curdircolor=0x27;
	if(FileRequest(&f,"gp:\\gpmm\\coleco",1)) GpAppExit();
	//error=GpRelativePathSet("gp:\\gpmm\\coleco"); if(error)	printf(0,0,"relpath error");
	//strcpy(file,"antarcti.rom");
	UPeriod=2;
	InitMachine();
	Verbose=0;
	StartColeco((char*)file);
	TrashColeco();
	TrashMachine();
	GpAppExit();
}

void Sound(int ch, int freq, int vol)
{
	//sprintf(buf,"%d: %d Hz at %d",ch,freq,vol); print(0,0,buf);
	freqs[ch]=freq;
	vols[ch]=vol;
	//int x=GpTickCountGet(); while(GpTickCountGet()<x+100);
}

void SoundCallback(void* userdata, unsigned char* stream, int len)
{
	int copied=0;
	while(copied<len && rbfill)
	{
		*stream++=RingBuf[rbrpos];
		rbrpos=(rbrpos+1)&(RBUF_SIZE-1);
		rbfill--;
		copied++;
	}
}
